import abc
from typing import Callable, Optional

import numpy as np
from scipy.special import expit, logit

from caliber.binary_classification.minimizing.base import (
    MinimizingBinaryClassificationModel,
)


class LinearScalingBinaryClassificationModel(
    MinimizingBinaryClassificationModel, abc.ABC
):
    def __init__(
        self,
        loss_fn: Callable[[np.ndarray, np.ndarray], float],
        threshold: float,
        minimize_options: Optional[dict] = None,
        has_intercept: bool = True,
        has_bivariate_slope: bool = False,
    ):
        self._has_intercept = has_intercept
        self._has_bivariate_slope = has_bivariate_slope
        super().__init__(loss_fn, threshold, minimize_options)

    def _predict_proba(self, params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        probs = np.clip(probs, 1e-6, 1 - 1e-6)
        if self._has_intercept:
            if self._has_bivariate_slope:
                return expit(
                    params[0]
                    + params[1] * np.log(probs)
                    - params[2] * np.log(1 - probs)
                )
            return expit(params[0] + params[1] * logit(probs))
        if self._has_bivariate_slope:
            return expit(params[0] * np.log(probs) - params[1] * np.log(1 - probs))
        return expit(params * logit(probs))



from functools import partial
from typing import Optional

from caliber.binary_classification.metrics.knee_distance import knee_point_distance
from caliber.binary_classification.minimizing.linear_scaling.calibration.base import (
    CalibrationLinearScalingBinaryClassificationModel,
)
from caliber.binary_classification.minimizing.linear_scaling.mixins.fit.brute_fit import (
    LinearScalingBruteFitBinaryClassificationMixin,
)


class KneePointLinearScalingBinaryClassificationModel(
    LinearScalingBruteFitBinaryClassificationMixin,
    CalibrationLinearScalingBinaryClassificationModel,
):
    def __init__(
        self,
        minimize_options: Optional[dict] = None,
        lam: float = 0.01,
        has_intercept: bool = True,
        has_bivariate_slope: bool = False,
        n_thresholds: int = 100,
    ):
        super().__init__(
            loss_fn=partial(knee_point_distance, n_thresholds=n_thresholds),
            minimize_options=minimize_options,
            has_intercept=has_intercept,
            has_bivariate_slope=has_bivariate_slope,
        )
        self._lam = lam



from typing import Literal

import numpy as np


def prediction_interval_coverage_probability(
    targets: np.ndarray,
    bounds: np.ndarray,
    interval_type: Literal["two-tailed", "left-tailed", "right-tailed"],
) -> np.ndarray:
    if targets.ndim > 1:
        raise ValueError("`targets` must be a 1-dimensional array.")
    if interval_type == "two-tailed":
        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(
                "`bounds` must be a two-dimensional array of bounds with `bounds.shape[1]` equal 2."
            )
        conds = (targets <= bounds[:, 1]) * (targets >= bounds[:, 0])
    elif interval_type == "left-tailed":
        if bounds.ndim != 1:
            raise ValueError(
                "`bounds` must be a one-dimensional array of right quantiles."
            )
        conds = targets <= bounds
    elif interval_type == "right-tailed":
        if bounds.ndim != 1:
            raise ValueError(
                "`bounds` must be a one-dimensional array of left quantiles."
            )
        conds = targets >= bounds
    else:
        raise ValueError(f"`interval_type={interval_type}` not recognized.")
    return np.mean(conds)



from typing import List

import numpy as np

from caliber.binary_classification.minimizing.mixins.fit.smooth_fit import (
    SmoothFitBinaryClassificationMixin,
)


class LinearScalingSmoothFitBinaryClassificationMixin(
    SmoothFitBinaryClassificationMixin
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "bounds" not in self._minimize_options:
            self._minimize_options["bounds"] = self._get_bounds()

    def _get_x0(self) -> np.ndarray:
        if self._has_intercept:
            if self._has_bivariate_slope:
                return np.array([0.0, 1.0, 1.0])
            return np.array([0.0, 1.0])
        if self._has_bivariate_slope:
            return np.array([1.0, 1.0])
        return np.array(1.0)

    def _get_bounds(self) -> List:
        if self._has_intercept:
            if self._has_bivariate_slope:
                return [(None, None), (0.0, None), (0.0, None)]
            return [(None, None), (0.0, None)]
        if self._has_bivariate_slope:
            return [(0.0, None), (0.0, None)]
        return [(0.0, None)]



import numpy as np


class BinaryClassificationChecksMixin:
    @staticmethod
    def _check_targets(targets: np.ndarray) -> None:
        if set(targets) not in [{0, 1}, {0}, {1}]:
            raise ValueError("`targets` must must include only values in {0, 1}.")
        if targets.ndim != 1:
            raise ValueError("`targets` must be a one-dimensional array.")

    @staticmethod
    def _check_probs(probs: np.ndarray) -> None:
        if probs.ndim != 1:
            raise ValueError("`probs` must be a one-dimensional array.")
        if np.sum(probs > 1) or np.sum(probs < 0):
            raise ValueError("All values in `probs` must be between 0 and 1.")



from typing import List, Tuple

from caliber.binary_classification.minimizing.linear_scaling.mixins.regularizers.l2 import (
    L2RegularizerMixin,
)
from caliber.binary_classification.minimizing.mixins.fit.global_fit import (
    GlobalFitBinaryClassificationMixin,
)


class LinearScalingGlobalFitBinaryClassificationMixin(
    L2RegularizerMixin, GlobalFitBinaryClassificationMixin
):
    def _get_bounds(self) -> List[Tuple]:
        if self._has_intercept:
            if self._has_bivariate_slope:
                return [(-3, 3), (0, 4), (0, 4)]
            return [(-3, 3), (0, 4)]
        if self._has_bivariate_slope:
            return [(0, 4), (0, 4)]
        return [(0, 4)]



from typing import Literal

import numpy as np

from caliber.regression.conformal_regression.base import (
    ConformalizedScoreRegressionModel,
)
from caliber.utils.interval_type_error import interval_type_error
from caliber.utils.quantile_checks import (
    left_tailed_quantile_check,
    right_tailed_quantile_check,
    two_tailed_quantile_check,
)


class ConformalizedQuantileRegressionModel(ConformalizedScoreRegressionModel):
    def __init__(
        self,
        confidence: float,
        interval_type: Literal[
            "two-tailed", "left-tailed", "right-tailed"
        ] = "two-tailed",
    ):
        super().__init__(confidence=confidence)
        self.interval_type = interval_type

    def fit(self, quantiles: np.ndarray, targets: np.ndarray) -> None:
        if self.interval_type == "two-tailed":
            two_tailed_quantile_check(quantiles)
            scores = np.maximum(quantiles[:, 0] - targets, targets - quantiles[:, 1])
        elif self.interval_type == "left-tailed":
            left_tailed_quantile_check(quantiles)
            scores = quantiles - targets
        elif self.interval_type == "right-tailed":
            right_tailed_quantile_check(quantiles)
            scores = targets - quantiles
        else:
            interval_type_error(self.interval_type)
        super().fit(scores, targets)

    def predict(self, quantiles: np.ndarray) -> np.ndarray:
        if self.interval_type == "two-tailed":
            two_tailed_quantile_check(quantiles)
            lowers = quantiles[:, 0] - self._params
            uppers = quantiles[:, 1] + self._params
            bounds = np.stack((lowers, uppers), axis=1)
        elif self.interval_type == "left-tailed":
            left_tailed_quantile_check(quantiles)
            bounds = quantiles - self._params
        elif self.interval_type == "right-tailed":
            right_tailed_quantile_check(quantiles)
            bounds = quantiles + self._params
        else:
            interval_type_error(self.interval_type)
        return bounds
